/* $Copyright$
 *  Copyright (C) Fujitsu Siemens Computers GmbH 2001, 2002
 *  All rights reserved
 */
#ident "$Header$"
#ifndef WITHOUT_INCLUDES
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE_SUB
#  define MODULE
#endif
#ifndef PRECOMPILE
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/pci.h>
#include <asm/io.h>             /* inb */
#include <asm/uaccess.h>        /* put_user */
#endif
#ident "$Header: $"
#ident "$Header$"
#ident "$Header: $"
struct i960_Pager
{
    unsigned long len;
    unsigned char *buffer;
    unsigned int info;
};
EXPORT_NO_SYMBOLS;
static const int SUCCESS = 0;
static const char *I960_DEV = "i960";
static const unsigned int VENDOR_ID = 0x8086;
static const unsigned int I960_DEVICE_ID = 0x0960;
static const unsigned int PII4X_DEVICE_ID = 0x7113;
static const unsigned long MAX_MEM_SIZE = 0x004000;
static const unsigned short PAGER_START = 0x800;
static const unsigned long MAX_PAGER_SIZE = 0x00010000;
static short i960Debug = 0;
MODULE_PARM(i960Debug, "h");
static int g_deviceOpen = -1;
static int g_deviceMap = -1;
static int g_i960_major = 0;
static struct pci_dev *g_i960Dev = NULL;
static u16 g_PII4X_port;
static unsigned long g_PII4X_physStartAddr;
static int readPager(unsigned long arg) {
        unsigned short PagerAddr = PAGER_START;
        unsigned char *MSBPagerAddr = ((unsigned char *)&PagerAddr)+1;
        unsigned char *LSBPagerAddr = (unsigned char *)&PagerAddr;
        struct i960_Pager *data = (struct i960_Pager *)arg;
        unsigned char *buf;
        unsigned long len;
        unsigned char tmp;
        int i;
        if (i960Debug) printk(KERN_DEBUG "i960 :" "-->readPager\n");
        if ( (inb(g_PII4X_port + 1) & 0x80) != 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "g_PII4X_port in use\n");
                return -EIO;
        }
        if (get_user(len, (unsigned long *)&(data->len)) < 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "get_user len failed\n");
                return -EFAULT;
        } else if (len > MAX_PAGER_SIZE) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "len invalid\n");
                return -EFAULT;
        } else if (get_user(buf, (unsigned char **)&data->buffer) < 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "get_user buf failed\n");
                return -EFAULT;
        }
        for (i=0; i < len; i++, buf++, PagerAddr++) {
                outb((*MSBPagerAddr) | 0x80, g_PII4X_port + 1);
                outb(*LSBPagerAddr, g_PII4X_port);
                tmp = inb(g_PII4X_port+2);
                put_user(tmp, buf);
        }
    outb((*MSBPagerAddr) | 0x7F, g_PII4X_port+1);
        if (i960Debug) printk(KERN_DEBUG "i960 :" "<--readPager\n");
        return SUCCESS;
}
static int writePager(unsigned long arg) {
        unsigned short PagerAddr = PAGER_START;
        unsigned char *MSBPagerAddr = ((unsigned char *)&PagerAddr)+1;
        unsigned char *LSBPagerAddr = (unsigned char *)&PagerAddr;
        struct i960_Pager *data = (struct i960_Pager *)arg;
        unsigned char *buf;
        unsigned long len;
        unsigned char tmp;
        int i;
        if (i960Debug) printk(KERN_DEBUG "i960 :" "-->writePager\n");
        if ( (inb(g_PII4X_port + 1) & 0x80) != 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "g_PII4X_port in use\n");
                return -EIO;
        }
        if (get_user(len, (unsigned long *)&(data->len)) < 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "get_user len failed\n");
                return -EFAULT;
        } else if (len > MAX_PAGER_SIZE) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "len invalid\n");
                return -EFAULT;
        } else if (get_user(buf, (unsigned char **)&data->buffer) < 0) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "get_user buf failed\n");
                return -EFAULT;
        }
        for (i=0; i < len; i++, buf++, PagerAddr++) {
                outb((*MSBPagerAddr) | 0x80, g_PII4X_port + 1);
                outb(*LSBPagerAddr, g_PII4X_port);
                get_user(tmp, buf);
                outb(tmp, g_PII4X_port + 2);
        }
    outb((*MSBPagerAddr) | 0x7F, g_PII4X_port+1);
        if (i960Debug) printk(KERN_DEBUG "i960 :" "<--writePager\n");
        return SUCCESS;
}
static int i960_ioctl(
                                struct inode *inode,
                                struct file *file,
                                unsigned int cmd,
                                unsigned long arg )
{
        if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_ioctl(0x%x)\n",cmd);
        if (_IOC_TYPE(cmd) != 'i') {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "  type != magic\n");
                return -EINVAL;
        }
        if (_IOC_NR(cmd) > 4) {
                if (i960Debug) printk(KERN_DEBUG "i960 :" "  nr > maxnr\n");
                return -EINVAL;
        }
        switch (cmd) {
        case _IOW('i', 0, int):
                                printk(KERN_DEBUG "i960 : i960Debug old=%d", i960Debug);
                                if (get_user(i960Debug, (int *)arg) < 0) {
                                        printk(KERN_DEBUG " set failed\n");
                                        return -EFAULT;
                                }
                                printk(KERN_DEBUG " new=%d\n", i960Debug);
                                break;
        case _IOR('i', 1, struct i960_Pager):
                                return readPager(arg);
        case _IOW('i', 2, struct i960_Pager):
                                return writePager(arg);
        case _IO ('i', 3): {
                                int i = 10;
                                if (i960Debug) printk(KERN_DEBUG "i960 :" " reset use count\n");
                                while(i-- > 0 && MOD_IN_USE) MOD_DEC_USE_COUNT;
                                if (i960Debug) printk(KERN_DEBUG "i960 :" "reset done.\n");
                                }
                                break;
        default:
                                if (i960Debug) printk(KERN_DEBUG "i960 :" " ioctl-cmd=%x not valid\n",cmd);
                                return -EINVAL;
        }
        return SUCCESS;
}
static int getPII4X_physicalStartAddress(void) {
    unsigned char byte;
    unsigned char *psa = (unsigned char *)&g_PII4X_physStartAddr;
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->getPII4X_physicalStartAddress\n");
    pci_read_config_word(g_i960Dev, 0x64, &g_PII4X_port);
    if (i960Debug) printk(KERN_DEBUG "i960 :" "Port=0x%x\n",g_PII4X_port);
    byte = inb(g_PII4X_port + 1) & 0x80;
    if (byte != 0) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "g_PII4X_port in use: byte=0x%x\n",byte);
        return 0;
    }
    outb(0x80, g_PII4X_port+1);
        outb(0x04, g_PII4X_port);
        *(psa) = inb(g_PII4X_port+2);
        outb(0x05, g_PII4X_port);
        *(psa + 1) = inb(g_PII4X_port+2);
        outb(0x06, g_PII4X_port);
        *(psa + 2) = inb(g_PII4X_port+2);
        outb(0x07, g_PII4X_port);
        *(psa + 3) = inb(g_PII4X_port+2);
    outb(0x7F, g_PII4X_port+1);
    if (i960Debug) printk(KERN_DEBUG "i960 :" "physStartAddr=0x%lx\n",g_PII4X_physStartAddr);
    g_PII4X_physStartAddr = (g_PII4X_physStartAddr & PAGE_MASK);
    if (i960Debug) printk(KERN_DEBUG "i960 :" "<--physStartAddr=0x%lx\n",g_PII4X_physStartAddr);
    return 1;
}
static int i960_open(struct inode *inode, struct file *file)
{
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_open: (%p,%p)\n",inode, file);
    if(g_deviceOpen < 0) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "i960_open: device not initialized\n");
        return -ENODEV;
    }
        if (g_deviceOpen > 0) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "i960_open: device is busy\n");
        return -EBUSY;
        }
    g_deviceOpen++;
    MOD_INC_USE_COUNT;
    return SUCCESS;
}
static int i960_release(struct inode *inode, struct file *file)
{
    g_deviceOpen--;
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_release: (%p,%p)\n",inode, file);
    MOD_DEC_USE_COUNT;
    return SUCCESS;
}
static void i960_vma_unmap(
                struct vm_area_struct *area,
                unsigned long addr,
                size_t len)
{
        g_deviceMap--;
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_vma_unmap\n");
}
static void i960_vma_open(struct vm_area_struct *area) {
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_vma_open\n");
    MOD_INC_USE_COUNT;
}
static void i960_vma_close(struct vm_area_struct *area) {
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_vma_close\n");
    i960_vma_unmap(area, area->vm_start, (size_t)(area->vm_end - area->vm_start));
    MOD_DEC_USE_COUNT;
}
static struct vm_operations_struct i960_vm_ops = {
    i960_vma_open,
    i960_vma_close,
};
static int i960_mmap(struct file *file, struct vm_area_struct *vma)
{
    int err;
    unsigned long virt_addr = vma->vm_start;
    unsigned long vsize = vma->vm_end - vma->vm_start;
    if (i960Debug) printk(KERN_DEBUG "i960 :" "-->i960_mmap: virt=0x%lx phys=0x%lx vsize=0x%lx\n",virt_addr, g_PII4X_physStartAddr, vsize);
#if defined(REDHAT) && defined(ADVANCED) 
    err = remap_page_range( 
	    vma, 
            virt_addr, 
            g_PII4X_physStartAddr, 
            vsize, 
            vma->vm_page_prot); 
#else
    err = remap_page_range(
            virt_addr,
            g_PII4X_physStartAddr,
            vsize,
            vma->vm_page_prot);
#endif
    if (err) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "i960_mmap: remap_page_range failed: %d\n",err);
        return -EINVAL;
    }
    if (vma->vm_ops) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "i960_mmap: vma->vm_ops != NULL\n");
        return -EINVAL;
    }
    vma->vm_ops = &i960_vm_ops;
        g_deviceMap++;
    MOD_INC_USE_COUNT;
    if (i960Debug) printk(KERN_DEBUG "i960 :" "<--i960_mmap\n");
    return SUCCESS;
}
static struct file_operations i960_fops =
{
    .ioctl = i960_ioctl,
    .mmap = i960_mmap,
    .open = i960_open,
    .release = i960_release
};
int init_module(void)
{
    if (! pci_present()) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "init_module: pcibios not present\n");
        return -ENODEV;
    }
    g_i960Dev = pci_find_device( VENDOR_ID, I960_DEVICE_ID, NULL);
    if (g_i960Dev == NULL) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "init_module: i960 not present!\n");
        return -ENODEV;
    }
    g_i960Dev = pci_find_device( VENDOR_ID, PII4X_DEVICE_ID, NULL);
    if (g_i960Dev == NULL) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "init_module: PII4X not present!\n");
        return -ENODEV;
    }
    if (getPII4X_physicalStartAddress() == 0) {
        if (i960Debug) printk(KERN_DEBUG "i960 :" "init_module: PII4X get physical start address failed!\n");
        return -EIO;
    }
    if((g_i960_major=register_chrdev(g_i960_major, I960_DEV, &i960_fops)) < 0) {
        printk(KERN_ERR "i960: unable to get major %d\n", g_i960_major);
        return -EIO;
    }
    g_deviceOpen = 0;
    return SUCCESS;
}
void cleanup_module(void)
{
    int err;
    err = unregister_chrdev(g_i960_major, I960_DEV);
    if (err < 0) {
        printk(KERN_ERR "i960: unregister_chrdev(%i) failed!\n", g_i960_major);
    }
    g_i960Dev = NULL;
    g_deviceOpen = -1;
    g_deviceMap = -1;
}